class cached_property(object):
    """
    This is a direct copy-paste of Django's cached property from
    https://github.com/django/django/blob/2456ffa42c33d63b54579eae0f5b9cf2a8cd3714/django/utils/functional.py#L38-50
    """
    def __init__(self, func):
        self.func = func

    def __get__(self, instance, type=None):
        if instance is None:
            return self
        res = instance.__dict__[self.func.__name__] = self.func(instance)
        return res


def print_temp(*args, file_name='/tmp/gistfinder.log'):
    with open(file_name, 'a') as buff:
        print(*args, file=buff)



import os
import sys
import dataset
import json
from .utils import cached_property

CONFIG_DIR = os.path.realpath(os.path.expanduser('~/.config/gistfinder'))

BASE_API_URL_ENV_KEY = 'GISTFINDER_API_URL'
BASE_URL_ENV_KEY = 'GISTFINDER_URL'


class Config:
    config_dir = CONFIG_DIR
    db_file = os.path.join(config_dir, 'database.sqlite')
    auth_file = os.path.join(config_dir, 'github_auth.json')

    db_url = 'sqlite:///{db_file}'.format(db_file=db_file)

    BASE_URL = 'https://gist.github.com'
    BASE_URL = os.environ.get(BASE_URL_ENV_KEY, BASE_URL)

    BASE_API_URL = 'https://api.github.com/users'
    BASE_API_URL = os.environ.get(BASE_API_URL_ENV_KEY, BASE_API_URL)

    LIST_TABLE = 'list_table'
    CODE_TABLE = 'code_table'

    def _check_okay(self):
        has_user = False
        has_token = False
        has_auth_file = os.path.isfile(self.auth_file)

        if has_auth_file:
            with open(self.auth_file) as buff:
                blob = json.load(buff)
                if blob.get('GIST_USER'):
                    has_user = True
                if blob.get('GIST_TOKEN'):
                    has_token = True
        
        msg = ''
        if not has_user:
            msg += '\nYou need to set up a user: gistfinder -u <username>\n'

        if not has_token:
            msg += '\nYou need to set up a token: gistfinder -t <access_token>'

        if not has_user and not has_token:
            msg = '\nYou need to set up a user and token: gistfinder -u <username> -t <access_token>'

        if msg:
            print(msg, file=sys.stderr)
            print('\n', file=sys.stdout)
            sys.exit(1)

    @cached_property
    def gist_user(self):
        self._check_okay()

        with open(self.auth_file) as f:
            user = json.loads(f.read()).get('GIST_USER')

        return user

    @cached_property
    def user_url(self):
        url = f'{self.BASE_API_URL.strip("/")}/{self.gist_user}/gists'
        return url

    @cached_property
    def db(self):
        self._check_okay()
        return dataset.connect(url=self.db_url)

    @property
    def list_table(self):
        return self.db[self.LIST_TABLE]

    @property
    def code_table(self):
        return self.db[self.CODE_TABLE]

    @cached_property
    def github_token(self):
        self._check_okay()

        with open(self.auth_file) as f:
            access_token = json.loads(f.read()).get('GIST_TOKEN')

        return access_token

    def update(self, **updates):
        blob = {}
        os.makedirs(self.config_dir, exist_ok=True)

        if os.path.isfile(self.auth_file):
            with open(self.auth_file) as buff:
                saved_blob = json.load(buff)
            blob.update(saved_blob)

        blob.update(updates)

        with open(self.auth_file, 'w') as out_file:
            json.dump(blob, out_file)

    def set_github_token(self, token):
        self.update(GIST_TOKEN=token)

    def set_github_user(self, user):
        self.update(GIST_USER=user)



from collections import OrderedDict
from fnmatch import fnmatch
from fuzzywuzzy import process
from .config import Config
from .utils import cached_property


class Loader(Config):
    @cached_property
    def has_tables(self):
        expected_tables = {Config.LIST_TABLE, Config.CODE_TABLE}
        existing_tables = set(self.db.tables)
        return existing_tables.intersection(expected_tables) == expected_tables

    @cached_property
    def records(self):
        if not self.has_tables:
            return OrderedDict()

        cursor = self.db.query(
            f"""
                select
                    l.gist_id AS gid,
                    l.description,
                    l.file AS file_name,
                    c.code
                from
                    {Config.LIST_TABLE} as l
                join
                    {Config.CODE_TABLE} as c
                on
                    c.file_url = l.file_url
                order by
                    file_name collate nocase asc,
                    description asc
            """
        )
        raw_records = list(cursor)

        out_recs = OrderedDict()
        for rec in raw_records:
            rec['text'] = '\n'.join(
                [
                    rec['file_name'],
                    rec['description'],
                    rec['code']
                ]
            )
            out_recs[rec['gid']] = rec
        return out_recs

    def rank(self, records, expr, field):
        search_recs = OrderedDict((t[0], t[1][field]) for t in records.items())
        tups = process.extract(expr, search_recs, limit=len(records) + 1)
        out = OrderedDict()
        for tup in tups:
            gid = tup[2]
            out[gid] = records[gid]
        return out

    def filter_glob(self, expr, records):
        out = OrderedDict()
        for gid, rec in records.items():
            if fnmatch(rec['file_name'], expr):
                out[gid] = rec
        return out

    def get(self, *, glob_expr=None, text_expr=None, desc_expr=None, file_expr=None, code_expr=None):
        records = self.records
        if glob_expr and records:
            records = self.filter_glob(glob_expr, records)

        if text_expr and records:
            records = self.rank(records, text_expr, 'text')

        if desc_expr and records:
            records = self.rank(records, desc_expr, 'description')

        if file_expr and records:
            records = self.rank(records, file_expr, 'file_name')

        if code_expr and records:
            records = self.rank(records, code_expr, 'code')

        return records



import os
import requests
import math
import time
from tqdm import tqdm
import sys
from .config import Config


class Updater(Config):
    PER_PAGE = 99
    MAX_GISTS = 10_000

    def blob_generator(self):
        for page in range(int(math.ceil(self.MAX_GISTS / self.PER_PAGE))):
            params = {
                'per_page': self.PER_PAGE,
                'page': page,
            }
            headers = {
                'Authorization': f'token {self.github_token}',
                'accept': 'application/vnd.github.v3+json'
            }
            resp = requests.get(self.user_url, params=params, headers=headers)
            self.latest_resp = resp
            blobs = resp.json()
            if blobs:
                yield blobs
            else:
                break

    def get_list(self):
        rec_list = []
        for ll in self.blob_generator():
            rec_list.extend(ll)

        return rec_list

    def get_parsed_blob_list(self):
        keys = [
            'url',
            'id',
            'description',
            'created_at',
            'updated_at',

        ]
        records = []

        for item in self.get_list():
            record_template = {k: item[k] for k in keys}
            for file in item['files'].values():
                record = record_template.copy()
                record['file'] = file['filename']
                record['language'] = file.get('language', 'unkown')
                record['file_url'] = file['raw_url']
                record['size'] = file.get('size', 0)
                record['gist_id'] = record.pop('id')
                records.append(record)

        dedup = {r['file_url']: r for r in records}
        records = list(dedup.values())
        records = [r for r in records if not r['file'].endswith('ipynb')]

        return records

    def update_list_table(self):
        recs = self.get_parsed_blob_list()
        table = self.list_table
        table.drop()
        table.insert_many(recs)
        table.create_index(['file_url'])

    def get_lookup_dict(self, table):
        return {r['file_url']: r for r in table.all()}

    def get_code(self, url):
        resp = requests.get(url)
        if resp.status_code == 200:
            return resp.text
        else:
            print(f'Problem retrieving {url}', file=sys.stderr)
            return None

    def update_code_table(self, verbose=True):
        #import pdb; pdb.set_trace()
        list_table = self.list_table
        code_table = self.code_table

        code_lookup = self.get_lookup_dict(code_table)
        list_lookup = self.get_lookup_dict(list_table)

        missing_file_urls = set(list_lookup.keys()) - set(code_lookup.keys())
        deleted_file_urls = set(code_lookup.keys()) - set(list_lookup.keys())

        code_table.delete(file_url={'in': tuple(deleted_file_urls)})

        if not missing_file_urls:
            return

        if verbose:
            missing_file_urls = tqdm(missing_file_urls)

        for file_url in missing_file_urls:
            code = self.get_code(file_url)
            if code:
                code_table.insert(dict(file_url=file_url, code=code))
                time.sleep(.1)

        code_table.create_index(['file_url'])

    def reset(self):
        if os.path.isfile(self.db_file):
            print('Removing {}'.format(self.db_file), file=sys.stderr)
            os.unlink(self.db_file)
        self.sync()

    def sync(self):
        self.update_list_table()
        self.update_code_table()
        print('\nSync Complete!', file=sys.stderr)



import warnings
warnings.filterwarnings("ignore")

import re
import sys

from prompt_toolkit import Application
from prompt_toolkit.buffer import Buffer
from prompt_toolkit.enums import EditingMode
from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.layout.containers import VSplit, Window, HSplit
from prompt_toolkit.layout.controls import BufferControl


from prompt_toolkit.layout.layout import Layout
from prompt_toolkit.filters import to_filter, Condition
from prompt_toolkit.styles import Style
from pygments.lexers import Python3Lexer

from prompt_toolkit.layout import NumberedMargin
from prompt_toolkit.lexers import PygmentsLexer
import click

from gistfinder.sync import Updater
from .loader import Loader
#  from .utils import print_temp

from prompt_toolkit.application.current import get_app


@Condition
def not_in_search_mode():
    app = get_app()
    return app.state.layout.current_window != app.state.search_window


class AppState:
    SEARCH_DEFAULT_TEXT = r' Search:/  Window:<space> Select:<enter> Exit:<ctrl-c> '

    def __init__(self, loader):
        self.glob_expr = None
        self.text_expr = None
        self.desc_expr = None
        self.file_expr = None
        self.code_expr = None

        self.loader = loader

        self.list_buffer = Buffer(on_cursor_position_changed=self.list_row_change)  # Editable buffer.
        # self.list_buffer.text = '\n'.join(self.list_lines)
        self.sync_list_lines()

        self.list_buffer.read_only = to_filter(True)
        self.list_buffer.app_state = self

        self.content_buffer = Buffer()  # Editable buffer.
        self.content_buffer.text = self.code(0)
        self.content_buffer.read_only = to_filter(True)
        self.content_buffer.app_state = self

        self.search_buffer = Buffer(on_text_changed=self.search_text_change)
        self.search_buffer.app_state = self
        self.search_buffer.read_only = to_filter(True)

        self.description_buffer = Buffer()
        self.description_buffer.app_state = self
        self.description_buffer.read_only = to_filter(True)

        self.slash_buffer = Buffer()
        self.slash_buffer.app_state = self
        self.slash_buffer.read_only = to_filter(True)

        self._index = 0
        self.print_on_exit = False

        self._list_lines = None

    def sync_list_lines(self):
        self.list_buffer.read_only = to_filter(False)
        self.list_buffer.text = '\n'.join(self.list_lines)
        self.list_buffer.read_only = to_filter(True)

    def clear_searches(self):
        self.glob_expr = None
        self.text_expr = None
        self.desc_expr = None
        self.file_expr = None
        self.code_expr = None

    @property
    def list_recs(self):
        return self.loader.get(
            glob_expr=self.glob_expr,
            text_expr=self.text_expr,
            desc_expr=self.desc_expr,
            file_expr=self.file_expr,
            code_expr=self.code_expr
        )

    @property
    def list_lines(self):
        return [r['file_name'] for r in self.list_recs.values()]

    @property
    def descriptions(self):
        return [r['description'] for r in self.loader.records.values()]

    def search_text_change(self, buffer):
        # rex_glob = re.compile(r'\\g([^\\]+)')
        # rex_code = re.compile(r'\\c([^\\]+)')
        # rex_file = re.compile(r'\\f([^\\]+)')
        # rex_text = re.compile(r'\\t([^\\]+)')
        rex_text = re.compile(r'([^\\]+)')
        rex_slash = re.compile(r'\\\s*$')
        rex_space = re.compile(r'^\s*$')

        app_state = buffer.app_state

        query = buffer.text
        # m_glob = rex_glob.search(query)
        # m_code = rex_code.search(query)
        # m_file = rex_file.search(query)
        m_text = rex_text.search(query)
        m_slash = rex_slash.search(query)
        m_space = rex_space.search(query)

        self.clear_searches()

        if m_slash or m_space:
            return

        # if m_glob:
        #     self.glob_expr = m_glob.group(1)
        # if m_code:
        #     self.code_expr = m_code.group(1)
        # if m_file:
        #     self.file_expr = m_file.group(1)
        # if m_text:
        #     self.text_expr = m_text.group(1)

        # if not any([bool(m) for m in [m_glob, m_code, m_file, m_text]]):
        # self.text_expr = query
        self.text_expr = m_text.group(1)

        app_state.sync_list_lines()
        self.set_code(0)
        self.set_description(0)
        return

    def list_row_change(self, buffer):
        doc = buffer.document
        pos = doc.cursor_position_row

        self.set_code(pos)
        self.set_description(pos)

    @property
    def select_gist_url(self):
        if self.list_recs:
            gid = list(self.list_recs.values())[self._index]['gid']
            return f'{self.loader.BASE_URL.strip("/")}/{self.loader.gist_user}/{gid}'
        else:
            return ''

    def code(self, index):
        self._index = index
        if self.list_recs:
            return list(self.list_recs.values())[self._index]['code']
        else:
            return ''

    def description(self, index):
        self._index = index
        if self.list_recs:
            return list(self.list_recs.values())[self._index]['description']
        else:
            return ''

    def set_code(self, index):
        content_buffer = self.content_buffer
        content_buffer.read_only = to_filter(False)
        content_buffer.text = self.code(index)
        content_buffer.read_only = to_filter(True)

    def set_description(self, index):
        description_buffer = self.description_buffer
        description_buffer.read_only = to_filter(False)
        description_buffer.text = f'   {self.description(index)}'
        description_buffer.read_only = to_filter(True)

    @property
    def selected_code(self):
        return self.code(self._index)

    @property
    def selected_file_name(self):
        return self.list_lines[self._index]

    @property
    def selected_description(self):
        return self.descriptions[self._index]

    def print(self):
        if not self.print_on_exit:
            return
        print('\n', file=sys.stderr)
        print(file=sys.stderr)
        print('#' + '=' * 75, file=sys.stderr)
        print(f'# {self.select_gist_url}')
        print('#' + '=' * 75, file=sys.stderr)
        print(self.selected_code, file=sys.stderr)
        print('', file=sys.stderr)

    def register_windows(self, *windows):
        self.windows = windows
        self.current_window_index = 0

    def focus_window(self, index):
        self.current_window_index = index
        return self.current_window

    def next_window(self):
        self.current_window_index = (self.current_window_index + 1) % len(self.windows)
        return self.current_window

    @property
    def current_window(self):
        return self.windows[self.current_window_index]


class UI:
    def __init__(self):
        loader = Loader()
        if not loader.has_tables:
            msg = 'You must run sync command'
            print(msg, file=sys.stderr)
            sys.exit(1)

        self.state = AppState(loader)

    def get_container(self):

        list_window = Window(
            width=55,
            left_margins=[NumberedMargin()],
            content=BufferControl(buffer=self.state.list_buffer, focusable=True),
            cursorline=True,
            style='bg:#AE9EC9 fg:black',
        )

        code_window = Window(
            left_margins=[NumberedMargin()],
            content=BufferControl(buffer=self.state.content_buffer, focusable=True, lexer=PygmentsLexer(Python3Lexer)),
            ignore_content_width=True
        )
        description_window = Window(
            content=BufferControl(
                buffer=self.state.description_buffer,
                focusable=False,
                # key_bindings=self.get_search_key_bindings(),
            ),
            height=1,
            style='bg:#1B2631  fg:#F1C40F',
        )

        search_window = Window(
            content=BufferControl(
                buffer=self.state.search_buffer,
                focusable=True,
                key_bindings=self.get_search_key_bindings(),
            ),
            height=1,
            style='bg:#1B2631  fg:#F1C40F',
        )

        slash_window = Window(
            content=BufferControl(
                buffer=self.state.slash_buffer,
                focusable=False,
            ),
            height=1,
            width=1,
            style='bg:#1B2631  fg:#F1C40F',
        )

        self.state.register_windows(list_window, code_window)
        self.state.search_window = search_window

        main_container = VSplit([list_window, code_window])
        search_container = VSplit([slash_window, search_window])

        root_container = HSplit([
            description_window,
            main_container,
            search_container,
        ])
        return root_container

    def get_search_key_bindings(self):
        kb = KeyBindings()

        @kb.add('enter', eager=True)
        def _(event):
            window_to_focus = event.app.state.focus_window(0)
            event.app.layout.focus(window_to_focus)
            event.app.state.sync_list_lines()

        return kb

    def get_key_bindings(self):
        kb = KeyBindings()

        @kb.add('c-c')
        def _(event):
            " Quit application. "
            event.app.exit()

        @kb.add('c-q')
        def _(event):
            " Quit application. "
            event.app.exit()

        @kb.add('enter')
        def _(event):
            " Quit application. "
            event.app.state.print_on_exit = True
            event.app.exit()

        @kb.add('space', filter=not_in_search_mode)
        def _(event):
            " Quit application. "
            window_to_focus = event.app.state.next_window()
            event.app.layout.focus(window_to_focus)

        @kb.add('/')
        def _(event):
            " Go into search mode "
            window_to_focus = event.app.state.search_window
            event.app.layout.focus(window_to_focus)
            event.app.state.search_buffer.read_only = to_filter(False)
            event.app.state.search_buffer.text = ''

            event.app.state.slash_buffer.read_only = to_filter(False)
            event.app.state.slash_buffer.text = '/'
            event.app.state.slash_buffer.read_only = to_filter(True)

        @kb.add('escape')
        def _(event):
            " Get out of search mode "
            window_to_focus = event.app.state.focus_window(0)
            event.app.layout.focus(window_to_focus)
            event.app.state.search_buffer.text = AppState.SEARCH_DEFAULT_TEXT
            event.app.state.search_buffer.read_only = to_filter(True)

            event.app.state.slash_buffer.read_only = to_filter(False)
            event.app.state.slash_buffer.text = ''
            event.app.state.slash_buffer.read_only = to_filter(True)

            event.app.state.clear_searches()
            event.app.state.sync_list_lines()

        return kb

    def run(self):
        root_container = self.get_container()
        kb = self.get_key_bindings()

        layout = Layout(root_container)
        style = Style(
            [
                ('cursor-line', 'fg:ansiwhite bg:#003366'),
                ('cursor-line', 'fg:#CCCCCC bg:#003366'),
            ]
        )

        app = Application(
            layout=layout,
            full_screen=True,
            key_bindings=kb,
            editing_mode=EditingMode.VI,
            mouse_support=True,
            style=style,
        )
        self.state.app = app
        self.state.layout = layout
        app.state = self.state

        import logging
        logger = logging.getLogger()
        logger.setLevel(logging.CRITICAL)

        app.run()
        self.state.print()


@click.command(help='A CLI tool for searching your gists')
@click.option('-u', '--user', help='Set up github user')
@click.option('-t', '--token', help='Set up github token')
@click.option('-s', '--sync', is_flag=True, help='Sync updated gists')
@click.option('-r', '--reset', is_flag=True, help='Delete and resync all gists')
def cli(user, token, sync, reset):
    updater = Updater()
    if reset:
        updater.reset()
    elif sync:
        updater.sync()
    elif token or user:
        if token:
            updater.set_github_token(token)
        if user:
            updater.set_github_user(user)

        print()
        print('Wrote config to: {}'.format(updater.auth_file))
        print()

    else:
        UI().run()



# import multiprocessing to avoid this bug (http://bugs.python.org/issue15881#msg170215)
import multiprocessing
assert multiprocessing
import re
from setuptools import setup, find_packages


def get_version():
    """
    Extracts the version number from the version.py file.
    """
    VERSION_FILE = 'gistfinder/version.py'
    mo = re.search(r'^__version__ = [\'"]([^\'"]*)[\'"]', open(VERSION_FILE, 'rt').read(), re.M)
    if mo:
        return mo.group(1)
    else:
        raise RuntimeError('Unable to find version string in {0}.'.format(VERSION_FILE))


install_requires = [
    'Pygments',
    'click',
    'dataset',
    'fuzzywuzzy',
    'prompt_toolkit',
    'python-Levenshtein',
    'pytz',
    'requests',
    'tqdm',
]

tests_require = [
    'coverage',
    'flake8',
    'mock',
    'nose',
    'wheel',
]

docs_require = [
    'Sphinx',
    'sphinx_rtd_theme'
]

extras_require = {
    'dev': tests_require + docs_require,
}

setup(
    name='gistfinder',
    version=get_version(),
    description='A CLI tool for searching your gists',
    long_description='A CLI tool for searching your gists',
    url='https://github.com/robdmc/gistfinder',
    author='Rob deCarvalho',
    author_email='unlisted@unlisted.net',
    keywords='',
    packages=find_packages(),
    classifiers=[
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    license='MIT',
    include_package_data=True,
    test_suite='nose.collector',
    install_requires=install_requires,
    tests_require=tests_require,
    extras_require=extras_require,
    zip_safe=False,
    entry_points={
        'console_scripts': [
            'gistfinder = gistfinder.console:cli',
            'gf = gistfinder.console:cli',
        ],
    }
)
